from typing import TypedDict

from hatch_cython.config import Config
from hatch_cython.types import ListStr, ListT
from hatch_cython.utils import options_kws


class ExtensionArg(TypedDict):
    name: str
    files: ListStr


def setup_py(
    *files: ListT[ListStr],
    options: Config,
):
    code = """
from setuptools import Extension, setup
from Cython.Build import cythonize

INCLUDES = {includes!r}
EXTENSIONS = {ext_files!r}

if __name__ == "__main__":
    exts = [
        Extension(  ex.get("name"),
                    ex.get("files"),
                    extra_compile_args={compile_args!r},
                    extra_link_args={extra_link_args!r},
                    include_dirs=INCLUDES,
                    libraries={libs!r},
                    library_dirs={lib_dirs!r},
                    define_macros={define_macros!r},
                    {keywords}
        ) for ex in EXTENSIONS
    ]
    ext_modules = cythonize(
            exts,
            compiler_directives={directives!r},
            include_path=INCLUDES,
            {cython}
    )
    setup(ext_modules=ext_modules)
"""
    kwds = options_kws(options.compile_kwargs)
    cython = options_kws(options.cythonize_kwargs)
    return code.format(
        compile_args=options.compile_args_for_platform,
        extra_link_args=options.compile_links_for_platform,
        directives=options.directives,
        ext_files=files,
        keywords=kwds,
        cython=cython,
        includes=options.includes,
        libs=options.libraries,
        lib_dirs=options.library_dirs,
        define_macros=options.define_macros,
    ).strip()



from sys import version_info
from typing import Literal, TypeVar, Union

T = TypeVar("T")

vmaj = (version_info[0], version_info[1])
if vmaj >= (3, 10):
    from collections.abc import Callable
    from typing import ParamSpec

    TupleT = tuple
    DictT = dict
    ListT = list
else:
    from typing import Callable, Dict, List, Tuple  # noqa: UP035

    from typing_extensions import ParamSpec

    TupleT = Tuple  # noqa: UP006
    DictT = Dict  # noqa: UP006
    ListT = List  # noqa: UP006

P = ParamSpec("P")
ListStr = ListT[str]
UnionT = Union
CorePlatforms = Literal[
    "darwin",
    "linux",
    "windows",
]
CallableT = Callable



import os
from contextlib import contextmanager
from types import SimpleNamespace
from unittest.mock import patch


def true_if_eq(*vals):
    def inner(v: str, *extra):
        merge = [*vals, *extra]
        ok = any(v == val for val in merge)
        print(f"ok: {v} {ok} ", merge)  # noqa: T201
        return ok

    return inner


true_x64_mac = true_if_eq("/usr/local/include", "/usr/local/lib", "/usr/local/opt")
true_arm_mac = true_if_eq("/opt/homebrew/lib", "/opt/homebrew/include")


@contextmanager
def patch_path(arch: str, *extra: str):
    arches = {
        "x86_64": true_x64_mac,
        "arm64": true_arm_mac,
    }
    h = arches[arch]

    def wrap(path):
        return h(path, *extra)

    with patch("hatch_cython.config.config.path.exists", wrap):
        yield


@contextmanager
def arch_platform(arch: str, platform: str):
    def aarchgetter():
        return arch

    def platformgetter():
        return platform

    try:
        with patch("hatch_cython.utils.plat", platformgetter):
            with patch("hatch_cython.utils.plat", platformgetter):
                with patch("hatch_cython.config.platform.plat", platformgetter):
                    with patch("hatch_cython.plugin.plat", platformgetter):
                        with patch("hatch_cython.utils.aarch", aarchgetter):
                            with patch("hatch_cython.config.platform.aarch", aarchgetter):
                                yield
    finally:
        print(f"Clean {arch}-{platform}")  # noqa: T201
        del aarchgetter, platformgetter
        pass


@contextmanager
def pyversion(maj="3", min="10", p="0"):  # noqa: A002
    try:
        with patch("platform.python_version_tuple", lambda: (maj, min, p)):
            yield
    finally:
        pass


@contextmanager
def import_module(gets_include, gets_libraries=None, gets_library_dirs=None, some_setup_op=None):
    def get_import(name: str):
        print(f"patched {name}")  # noqa: T201
        return SimpleNamespace(
            gets_include=gets_include,
            gets_libraries=gets_libraries,
            gets_library_dirs=gets_library_dirs,
            some_setup_op=some_setup_op,
        )

    try:
        with patch(
            "hatch_cython.config.config.import_module",
            get_import,
        ):
            yield
    finally:
        pass


@contextmanager
def override_dir(path: str):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def override_env(d: dict):
    current = os.environ.copy()

    try:
        new = os.environ.copy()
        for k, v in d.items():
            new[k] = v
        os.environ.update(new)
        with patch(
            "hatch_cython.config.flags.environ",
            new,
        ):
            yield
    finally:
        for k, v in current.items():
            os.environ[k] = v



# distutils: language=c++

cimport numpy as cnp

from cython.parallel import parallel, prange

cnp.import_array()

ctypedef cnp.longlong_t dtype

cpdef str hello_world(str name):
    cdef str response
    response = f"hello, {name}"
    return response


cpdef dtype hello_numpy(cnp.ndarray[dtype, ndim=1] arr):
    cdef dtype tot
    cdef Py_ssize_t cap
    cdef int i
    tot = 0
    cap = arr.size
    with nogil, parallel():
        for i in prange(cap):
            with gil:
                tot += arr[i]
    return tot



# SPDX-FileCopyrightText: 2023-present joshua-auchincloss <joshua.auchincloss@proton.me>
#
# SPDX-License-Identifier: MIT
from os import getcwd, name
from sys import path

if name == "nt":
    from pathlib import WindowsPath

    p = WindowsPath(getcwd()) / "src"
else:
    from pathlib import PosixPath

    p = PosixPath(getcwd()) / "src"

path.append(str(p))

from hatch_cython.__about__ import __version__  # noqa: E402
from hatch_cython.devel import CythonBuildHook, src  # noqa: E402



import platform

from hatch_cython.constants import NORM_GLOB, UAST
from hatch_cython.types import CallableT, P, T


def memo(func: CallableT[P, T]) -> CallableT[P, T]:
    value = None
    ran = False

    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        nonlocal value, ran
        if not ran:
            value = func(*args, **kwargs)
            ran = True
        return value

    return wrapped


@memo
def plat():
    return platform.system().lower()


@memo
def aarch():
    return platform.machine().lower()


def options_kws(kwds: dict):
    return ",\n\t".join((f"{k}={v!r}" for k, v in kwds.items()))


def parse_user_glob(uglob: str):
    return uglob.replace("\\*", UAST).replace("*", NORM_GLOB).replace(UAST, "*")


@memo
def path_delim() -> str:
    if plat() == "windows":
        return ";"
    return ":"



from pytest import raises

from hatch_cython.config.macros import parse_macros


def test_macros():
    test_ok = [
        ["NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"],
        ["ABC_DEF"],
        ["GHI_JKL"],
    ]
    assert parse_macros(test_ok) == [
        ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
        ("ABC_DEF", None),
        ("GHI_JKL", None),
    ]


def test_invalid():
    test_not_ok = [
        # extra value
        ["a", "b", "c"]
    ]
    with raises(
        ValueError,
        match="macros must be defined",
    ):
        parse_macros(test_not_ok)



from hatch_cython.types import CorePlatforms, ListT

NORM_GLOB = r"([^\s]*)"
UAST = "${U_AST}"
EXIST_TRIM = 2
ANON = "anon"
INCLUDE = "include_"
OPTIMIZE = "-O2"
DIRECTIVES = {
    "binding": True,
    "language_level": 3,
}
LTPY311 = "python_version < '3.11'"
MUST_UNIQUE = ["-O", "-arch", "-march"]
POSIX_CORE: ListT[CorePlatforms] = ["darwin", "linux"]



# distutils: language=c++
cimport cython


@cython.final
cdef class ValueDefn:
    cdef public int value
    cpdef bint set(self, int value)


cdef inline int vmul(ValueDefn v1, ValueDefn v2):
    cdef int result
    result = v1.value ** v2.value
    return result



from textwrap import dedent
from types import SimpleNamespace

from toml import loads

from hatch_cython.config import parse_from_dict

from .utils import arch_platform, import_module, patch_path, pyversion


def test_config_parser():
    data = """
    [options]
    includes = []
    include_numpy = false
    include_pyarrow = false

    include_somelib = { pkg = "somelib", include = "gets_include", libraries = "gets_libraries", library_dirs = "gets_library_dirs", required_call = "some_setup_op" }

    compile_args = [
        { platforms = ["windows"], arg = "-std=c++17" },
        { platforms = ["linux", "darwin"], arg = "-I/abc/def" },
        { platforms = ["linux", "darwin"], arg = "-Wcpp" },
        { platforms = ["darwin"], arg = "-L/usr/local/opt/llvm/include" },
        { arch = ["anon"], arg = "-O1" },
        { arch = ["x86_64"], arg = "-O2" },
        { arch = ["arm64"], arg = "-O3" },
        { arg = "-py39", marker = "python_version == '3.9'" },
    ]
    extra_link_args =  [
        { platforms = ["darwin"],  arg = "-L/usr/local/opt/llvm/lib" },
        { platforms = ["windows"],  arg = "-LC://abc/def" },
        { platforms = ["linux"], arg = "-L/etc/ssl/ssl.h" },
        { arch = ["arm64"], arg = "-L/usr/include/cpu/simd.h" },
    ]

    define_macros = [
        ["NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"],
        ["ABC_DEF"],
        ["GHI_JKL"],
    ]
    directives = { boundscheck = false, nonecheck = false, language_level = 3, binding = true }

    abc_compile_kwarg = "test"
    """  # noqa: E501

    def gets_include():
        return "abc"

    def gets_libraries():
        return ["lib-a"]

    def gets_library_dirs():
        return ["dir-a"]

    ran = False

    def some_setup_op():
        nonlocal ran
        ran = True

    with pyversion():
        with import_module(
            gets_include=gets_include,
            gets_libraries=gets_libraries,
            gets_library_dirs=gets_library_dirs,
            some_setup_op=some_setup_op,
        ):

            def getcfg():
                parsed = loads(dedent(data))
                return parse_from_dict(SimpleNamespace(config=parsed))

            cfg = getcfg()
            assert ran
            assert len(getcfg().compile_args)

            with pyversion("3", "9"), arch_platform("arm64", "darwin"), patch_path("arm64"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-I/opt/homebrew/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-L/usr/local/opt/llvm/include",
                    "-py39",
                    "-O3",
                ]

            with arch_platform("x86_64", "windows"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-std=c++17",
                    "-O2",
                ]
                assert cfg.compile_links_for_platform == ["-LC://abc/def"]
            with arch_platform("x86_64", "linux"), patch_path("x86_64", "/usr/local/opt"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-I/usr/local/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-O2",
                ]
                assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt", "-L/etc/ssl/ssl.h"]
            with arch_platform("x86_64", "darwin"), patch_path("x86_64"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-I/usr/local/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-L/usr/local/opt/llvm/include",
                    "-O2",
                ]
                assert cfg.compile_links_for_platform == [
                    "-L/usr/local/lib",
                    "-L/usr/local/opt",
                    "-L/usr/local/opt/llvm/lib",
                ]

            with arch_platform("arm64", "windows"):
                cfg = getcfg()

                assert cfg.compile_args_for_platform == [
                    "-std=c++17",
                    "-O3",
                ]
                assert cfg.compile_links_for_platform == ["-LC://abc/def", "-L/usr/include/cpu/simd.h"]
            with arch_platform("arm64", "linux"), patch_path("x86_64"):
                cfg = getcfg()

                assert cfg.compile_args_for_platform == [
                    "-I/usr/local/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-O3",
                ]
                assert cfg.compile_links_for_platform == [
                    "-L/usr/local/lib",
                    "-L/usr/local/opt",
                    "-L/etc/ssl/ssl.h",
                    "-L/usr/include/cpu/simd.h",
                ]
            with arch_platform("arm64", "darwin"), patch_path("arm64"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-I/opt/homebrew/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-L/usr/local/opt/llvm/include",
                    "-O3",
                ]
                assert cfg.compile_links_for_platform == [
                    "-L/opt/homebrew/lib",
                    "-L/usr/local/opt/llvm/lib",
                    "-L/usr/include/cpu/simd.h",
                ]

            with arch_platform("", "windows"):
                cfg = getcfg()

                assert cfg.compile_args_for_platform == ["-std=c++17", "-O1"]
                assert cfg.compile_links_for_platform == [
                    "-LC://abc/def",
                ]
            with arch_platform("", "linux"), patch_path("x86_64"):
                cfg = getcfg()

                assert cfg.compile_args_for_platform == ["-I/usr/local/include", "-I/abc/def", "-Wcpp", "-O1"]
                assert cfg.compile_links_for_platform == [
                    "-L/usr/local/lib",
                    "-L/usr/local/opt",
                    "-L/etc/ssl/ssl.h",
                ]
            with arch_platform("", "darwin"), patch_path("x86_64"):
                cfg = getcfg()
                assert cfg.compile_args_for_platform == [
                    "-I/usr/local/include",
                    "-I/abc/def",
                    "-Wcpp",
                    "-L/usr/local/opt/llvm/include",
                    "-O1",
                ]
                assert cfg.compile_links_for_platform == [
                    "-L/usr/local/lib",
                    "-L/usr/local/opt",
                    "-L/usr/local/opt/llvm/lib",
                ]

            cfg = getcfg()

    assert cfg.directives == {"boundscheck": False, "nonecheck": False, "language_level": 3, "binding": True}
    assert cfg.libraries == gets_libraries()
    assert cfg.library_dirs == gets_library_dirs()
    assert gets_include() in cfg.includes
    assert cfg.compile_kwargs == {"abc_compile_kwarg": "test"}


def test_defaults():
    data = """
    [options]
    """
    parsed = loads(dedent(data))

    def getcfg():
        return parse_from_dict(SimpleNamespace(config=parsed))

    cfg = getcfg()
    assert cfg.directives == {"language_level": 3, "binding": True}

    with arch_platform("x86_64", "windows"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == [
            "-O2",
        ]
        assert cfg.compile_links_for_platform == []
    with arch_platform("x86_64", "linux"), patch_path("x86_64"):
        cfg = getcfg()
        assert cfg.compile_args_for_platform == [
            "-I/usr/local/include",
            "-O2",
        ]
        assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt"]
    with arch_platform("x86_64", "darwin"), patch_path("x86_64"):
        cfg = getcfg()
        assert cfg.compile_args_for_platform == [
            "-I/usr/local/include",
            "-O2",
        ]
        assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt"]

    with arch_platform("arm64", "windows"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == [
            "-O2",
        ]
        assert cfg.compile_links_for_platform == []
    with arch_platform("arm64", "linux"), patch_path("x86_64"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == [
            "-I/usr/local/include",
            "-O2",
        ]
        assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt"]
    with arch_platform("arm64", "darwin"), patch_path("arm64"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == [
            "-I/opt/homebrew/include",
            "-O2",
        ]
        assert cfg.compile_links_for_platform == [
            "-L/opt/homebrew/lib",
        ]

    with arch_platform("", "windows"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == ["-O2"]
        assert cfg.compile_links_for_platform == []

    with arch_platform("", "linux"), patch_path("x86_64", "/etc/ssl/ssl.h"):
        cfg = getcfg()

        assert cfg.compile_args_for_platform == ["-I/usr/local/include", "-O2"]
        assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt"]

    with arch_platform("", "darwin"), patch_path("x86_64"):
        cfg = getcfg()
        assert cfg.compile_args_for_platform == [
            "-I/usr/local/include",
            "-O2",
        ]
        assert cfg.compile_links_for_platform == ["-L/usr/local/lib", "-L/usr/local/opt"]

    cfg = getcfg()
    assert cfg.compile_kwargs == {}



import ast
from collections.abc import Generator
from unittest.mock import patch

from hatch_cython.config import Config, PlatformArgs
from hatch_cython.plugin import setup_py

from .utils import arch_platform, true_if_eq


def clean(s: str):
    return "\n".join(v.strip() for v in s.splitlines() if v.strip() != "")


EXPECT = """
from setuptools import Extension, setup
from Cython.Build import cythonize

INCLUDES = ['/123']
EXTENSIONS = ({'name': 'abc.def', 'files': ['./abc/def.pyx']}, {'name': 'abc.depb', 'files': ['./abc/depb.py']})

if __name__ == "__main__":
    exts = [
        Extension(  ex.get("name"),
                    ex.get("files"),
                    extra_compile_args=['-O2'],
                    extra_link_args=['-I/etc/abc/linka.h'],
                    include_dirs=INCLUDES,
                    libraries=['/abc'],
                    library_dirs=['/def'],
                    define_macros=[],

        ) for ex in EXTENSIONS
    ]
    ext_modules = cythonize(
            exts,
            compiler_directives={'binding': True, 'language_level': 3},
            include_path=INCLUDES,
            abc='def'
    )
    setup(ext_modules=ext_modules)
""".strip()


def test_setup_py():
    cfg = Config(
        includes=["/123"],
        libraries=["/abc"],
        library_dirs=["/def"],
        cythonize_kwargs={"abc": "def"},
        extra_link_args=[PlatformArgs(arg="-I/etc/abc/linka.h")],
    )
    with patch("hatch_cython.config.config.path.exists", true_if_eq()):
        with arch_platform("x86_64", ""):
            setup = setup_py(
                {"name": "abc.def", "files": ["./abc/def.pyx"]},
                {"name": "abc.depb", "files": ["./abc/depb.py"]},
                options=cfg,
            )
    assert clean(setup) == clean(EXPECT)


def test_solo_ext_type_validations():
    cfg = Config(
        includes=["/123"],
        libraries=["/abc"],
        library_dirs=["/def"],
        cythonize_kwargs={"abc": "def"},
        extra_link_args=[PlatformArgs(arg="-I/etc/abc/linka.h")],
    )
    with patch("hatch_cython.config.config.path.exists", true_if_eq()):
        with arch_platform("x86_64", ""):
            setup = setup_py(
                {"name": "abc.def", "files": ["./abc/def.pyx"]},
                options=cfg,
            )
    tested = False
    exteq = "EXTENSIONS ="
    for ln in setup.splitlines():
        if ln.startswith(exteq):
            tested = True
            ext = ast.literal_eval(ln.replace(exteq, "").strip())
            assert isinstance(ext, (list, tuple, Generator))  # noqa: UP038
            for ex in ext:
                assert isinstance(ex, dict)
                assert isinstance(ex.get("name"), str)
                assert isinstance(ex.get("files"), list)

    if not tested:
        raise ValueError(setup, tested, "missed test")



from hatch_cython.types import CallableT, DictT, ListT, P, TupleT, UnionT


# basic test to assert we can use subscriptable generics
def test_type_compat():
    TupleT[int, str]
    DictT[str, str]
    ListT[str]
    CallableT[P, str]
    UnionT[str, None]



import shutil
from os import getcwd, path
from pathlib import Path  # noqa: F401

import pytest
from toml import load

from hatch_cython.plugin import CythonBuildHook

from .utils import arch_platform, override_dir


def join(*rel):
    return path.join(getcwd(), *rel)


def read(rel: str):
    return open(join(*rel.split("/"))).read()


@pytest.fixture
def new_proj(tmp_path):
    project_dir = tmp_path / "app"
    project_dir.mkdir()

    (project_dir / "pyproject.toml").write_text(read("example/pyproject.toml"))
    (project_dir / "hatch.toml").write_text(read("example/hatch.toml"))
    (project_dir / "README.md").write_text(read("example/README.md"))
    (project_dir / "LICENSE.txt").write_text(read("example/LICENSE.txt"))
    (project_dir / "bootstrap.py").write_text(read("example/bootstrap.py"))
    shutil.copytree(join("example", "src"), (project_dir / "src"))
    shutil.copytree(join("example", "tests"), (project_dir / "tests"))
    return project_dir


def test_build_hook(new_proj):
    hook = CythonBuildHook(
        new_proj,
        load(new_proj / "hatch.toml")["build"]["targets"]["wheel"]["hooks"]["custom"],
        {},
        {},
        new_proj,
        "abc",
    )

    assert hook.is_src

    with arch_platform("", "windows"):
        assert hook.is_windows

    with arch_platform("", "darwin"):
        assert not hook.is_windows

    with arch_platform("", "linux"):
        assert not hook.is_windows

    assert hook.dir_name == "example_lib"

    proj = "./src/example_lib"
    assert hook.project_dir == proj

    assert sorted(hook.precompiled_globs) == sorted(
        [
            "./src/example_lib/*.py",
            "./src/example_lib/**/*.py",
            "./src/example_lib/*.pyx",
            "./src/example_lib/**/*.pyx",
            "./src/example_lib/*.pxd",
            "./src/example_lib/**/*.pxd",
        ]
    )

    with override_dir(new_proj):
        assert sorted(hook.normalized_included_files) == sorted(
            [
                "./src/example_lib/__about__.py",
                "./src/example_lib/__about__.py",
                "./src/example_lib/__init__.py",
                "./src/example_lib/__init__.py",
                "./src/example_lib/_alias.pyx",
                "./src/example_lib/_alias.pyx",
                "./src/example_lib/mod_a/__init__.py",
                "./src/example_lib/mod_a/adds.pyx",
                "./src/example_lib/mod_a/deep_nest/creates.pyx",
                "./src/example_lib/mod_a/some_defn.pxd",
                "./src/example_lib/mod_a/some_defn.py",
                "./src/example_lib/normal.py",
                "./src/example_lib/normal.py",
                "./src/example_lib/test.pyx",
                "./src/example_lib/test.pyx",
            ]
        )

        assert sorted(
            [{**ls, "files": sorted(ls.get("files"))} for ls in hook.grouped_included_files],
            key=lambda x: x.get("name"),
        ) == [
            {
                "name": "example_lib.__about__",
                "files": ["./src/example_lib/__about__.py", "./src/example_lib/__about__.py"],
            },
            {
                "name": "example_lib.__init__",
                "files": ["./src/example_lib/__init__.py", "./src/example_lib/__init__.py"],
            },
            {"name": "example_lib.aliased", "files": ["./src/example_lib/_alias.pyx", "./src/example_lib/_alias.pyx"]},
            {"name": "example_lib.mod_a.__init__", "files": ["./src/example_lib/mod_a/__init__.py"]},
            {"name": "example_lib.mod_a.adds", "files": ["./src/example_lib/mod_a/adds.pyx"]},
            {"name": "example_lib.mod_a.deep_nest.creates", "files": ["./src/example_lib/mod_a/deep_nest/creates.pyx"]},
            {
                "name": "example_lib.mod_a.some_defn",
                "files": ["./src/example_lib/mod_a/some_defn.py", "./src/example_lib/mod_a/some_defn.py"],
            },
            {"name": "example_lib.normal", "files": ["./src/example_lib/normal.py", "./src/example_lib/normal.py"]},
            {"name": "example_lib.test", "files": ["./src/example_lib/test.pyx", "./src/example_lib/test.pyx"]},
        ]

        rf = sorted(
            [
                "./src/example_lib/__about__.*.pxd",
                "./src/example_lib/__about__.*.pxd",
                "./src/example_lib/__about__.*.py",
                "./src/example_lib/__about__.*.py",
                "./src/example_lib/__about__.*.pyx",
                "./src/example_lib/__about__.*.pyx",
                "./src/example_lib/__init__.*.pxd",
                "./src/example_lib/__init__.*.pxd",
                "./src/example_lib/__init__.*.py",
                "./src/example_lib/__init__.*.py",
                "./src/example_lib/__init__.*.pyx",
                "./src/example_lib/__init__.*.pyx",
                "./src/example_lib/_alias.*.pxd",
                "./src/example_lib/_alias.*.pxd",
                "./src/example_lib/_alias.*.py",
                "./src/example_lib/_alias.*.py",
                "./src/example_lib/_alias.*.pyx",
                "./src/example_lib/_alias.*.pyx",
                "./src/example_lib/mod_a/__init__.*.pxd",
                "./src/example_lib/mod_a/__init__.*.py",
                "./src/example_lib/mod_a/__init__.*.pyx",
                "./src/example_lib/mod_a/adds.*.pxd",
                "./src/example_lib/mod_a/adds.*.py",
                "./src/example_lib/mod_a/adds.*.pyx",
                "./src/example_lib/mod_a/deep_nest/creates.*.pxd",
                "./src/example_lib/mod_a/deep_nest/creates.*.py",
                "./src/example_lib/mod_a/deep_nest/creates.*.pyx",
                "./src/example_lib/mod_a/some_defn.*.pxd",
                "./src/example_lib/mod_a/some_defn.*.pxd",
                "./src/example_lib/mod_a/some_defn.*.py",
                "./src/example_lib/mod_a/some_defn.*.py",
                "./src/example_lib/mod_a/some_defn.*.pyx",
                "./src/example_lib/mod_a/some_defn.*.pyx",
                "./src/example_lib/normal.*.pxd",
                "./src/example_lib/normal.*.pxd",
                "./src/example_lib/normal.*.py",
                "./src/example_lib/normal.*.py",
                "./src/example_lib/normal.*.pyx",
                "./src/example_lib/normal.*.pyx",
                "./src/example_lib/test.*.pxd",
                "./src/example_lib/test.*.pxd",
                "./src/example_lib/test.*.py",
                "./src/example_lib/test.*.py",
                "./src/example_lib/test.*.pyx",
                "./src/example_lib/test.*.pyx",
            ]
        )
        assert sorted(hook.normalized_artifact_globs) == rf

        assert sorted(hook.artifact_patterns) == [f"/{f}" for f in rf]

        hook.clean([])

        build_data = {
            "artifacts": [],
            "force_include": {},
        }
        hook.initialize("0.1.0", build_data)

        assert build_data.get("infer_tag")
        assert not build_data.get("pure_python")
        assert sorted(build_data.get("artifacts")) == sorted([f"/{f}" for f in rf])
        assert len(build_data.get("force_include")) == 9



import re
from dataclasses import dataclass, field

from hatch_cython.types import DictT, ListStr, UnionT
from hatch_cython.utils import parse_user_glob


@dataclass
class FileArgs:
    exclude: ListStr = field(default_factory=list)
    aliases: DictT[str, str] = field(default_factory=dict)

    def __post_init__(self):
        rep = {}
        for k, v in self.aliases.items():
            rep[parse_user_glob(k)] = v
        self.aliases = rep

    def matches_alias(self, other: str) -> UnionT[str, None]:
        matched = [re.match(v, other) for v in self.aliases.keys()]
        if any(matched):
            first = 0
            for ok in matched:
                if ok:
                    break
                first += 1
            return self.aliases[list(self.aliases.keys())[first]]



from hatch_cython.types import ListT, TupleT, UnionT

DefineMacros = ListT[TupleT[str, UnionT[str, None]]]


def parse_macros(define: ListT[ListT[str]]) -> DefineMacros:
    """Parses define_macros from list[list[str, ...]] -> list[tuple[str, str|None]]

    Args:
        define (ListT[ListT[str]]): list of listed strings of len 1 or 2. raises error if len > 2

    Raises:
        ValueError: length > 2 or types are not valid

    Returns:
        DefineMacros: list[tuple[str,str|None]]
    """
    for i, inst in enumerate(define):
        size = len(inst)
        if not (isinstance(inst, list) and size in (1, 2) and all(isinstance(v, str) or v is None for v in inst)):
            msg = "".join(
                f"define_macros[{i}]: macros must be defined as [name, <value>], "
                "where None value denotes #define FOO"
            )
            raise ValueError(msg, inst)
        inst: list
        if size == 1:
            define[i] = (inst[0], None)
        else:
            define[i] = (inst[0], inst[1])
    return define



from textwrap import dedent
from types import SimpleNamespace

import toml

from hatch_cython.config import parse_from_dict

from .utils import arch_platform, override_env


def test_xenvvars():
    data = """
    [options]
    env = [
        { env = "CC", arg = "c++" },
        { env = "CPP", arg = "c++" },
        { env = "CXX", arg = "c++" },
        { env = "CFLAGS", arg = "flag2" },
        { env = "CPPFLAGS", arg = "flag2" },
        { env = "LDFLAGS", arg = "flag2" },
        { env = "CUSTOM_1", arg = "flag2",  arch = "x86_64" },
        { env = "CUSTOM_2", arg = "flag2",  merges = true },
    ]
    """

    def getcfg():
        parsed = toml.loads(dedent(data))
        return parse_from_dict(SimpleNamespace(config=parsed))

    f1 = "flag1"
    f2 = "flag2"
    with override_env({"CFLAGS": f1, "CPPFLAGS": f1, "LDFLAGS": f1}):
        cfg = getcfg()

        f12 = "flag1 flag2"
        assert cfg.envflags.CC.arg == "c++"
        assert cfg.envflags.LDFLAGS.arg == "flag2"

        assert cfg.envflags.env.get("CC") == "c++"
        assert cfg.envflags.env.get("CFLAGS") == f12
        assert cfg.envflags.env.get("CPPFLAGS") == f12
        assert cfg.envflags.env.get("LDFLAGS") == f12

        with override_env({"CUSTOM_1": f1}):
            with arch_platform("x86_64", ""):
                cfg = getcfg()
                assert cfg.envflags.env.get("CUSTOM_1") == f2
            with arch_platform("", ""):
                cfg = getcfg()
                assert cfg.envflags.env.get("CUSTOM_1") == f1

        with override_env({"CUSTOM_2": f1}):
            cfg = getcfg()
            assert cfg.envflags.env.get("CUSTOM_2") == f12



from hatch_cython.config.platform import PlatformArgs
from hatch_cython.constants import POSIX_CORE


def get_default_link():
    return [
        PlatformArgs(arg="-L/opt/homebrew/lib", platforms=POSIX_CORE, depends_path=True),
        PlatformArgs(arg="-L/usr/local/lib", platforms=POSIX_CORE, depends_path=True),
        PlatformArgs(arg="-L/usr/local/opt", platforms=POSIX_CORE, depends_path=True),
    ]


def get_default_compile():
    return [
        PlatformArgs(arg="-O2"),
        PlatformArgs(arg="-I/opt/homebrew/include", platforms=POSIX_CORE, depends_path=True),
        PlatformArgs(arg="-I/usr/local/include", platforms=POSIX_CORE, depends_path=True),
    ]



def test_muls():
    from example_lib.mod_a.adds import fmul, imul

    assert fmul(5.5, 5.5) == 30.25
    assert imul(21, 2) == 42


def test_vals():
    from example_lib.mod_a.some_defn import ValueDefn

    v = ValueDefn(10)
    assert v.value == 10
    v.set(5)
    assert v.value == 5


def test_deep_nesting():
    from example_lib.mod_a.deep_nest.creates import fast_create

    o = fast_create()
    assert o.do() == "abc"



from numpy import linspace, longlong

from example_lib.test import hello_numpy, hello_world


def test_hello_world():
    assert hello_world("world") == "hello, world"


def test_hello_numpy():
    assert hello_numpy(linspace(0, 100, 25, dtype=longlong)) == 1240



import os
import re
import subprocess
import sys
from contextlib import contextmanager
from glob import glob
from tempfile import TemporaryDirectory
from typing import ClassVar

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from hatch_cython.config import parse_from_dict
from hatch_cython.temp import ExtensionArg, setup_py
from hatch_cython.types import ListStr, ListT, P
from hatch_cython.utils import memo, parse_user_glob, plat


class CythonBuildHook(BuildHookInterface):
    PLUGIN_NAME = "cython"

    precompiled_extension: ClassVar[list] = [
        ".pyx",
        ".pxd",
    ]
    intermediate_extensions: ClassVar[list] = [
        ".c",
        ".cpp",
    ]
    compiled_extensions: ClassVar[list] = [
        ".dll",
        # unix
        ".so",
        # windows
        ".pyd",
    ]

    def __init__(self, *args: P.args, **kwargs: P.kwargs):
        super().__init__(*args, **kwargs)

    @property
    @memo
    def is_src(self):
        return os.path.exists(os.path.join(self.root, "src"))

    @property
    def is_windows(self):
        return plat() == "windows"

    def normalize_path(self, pattern: str):
        if self.is_windows:
            return pattern.replace("/", "\\")
        return pattern.replace("\\", "/")

    def normalize_glob(self, pattern: str):
        return pattern.replace("\\", "/")

    @property
    @memo
    def dir_name(self):
        return self.options.src if self.options.src is not None else self.metadata.name

    @property
    @memo
    def project_dir(self):
        if self.is_src:
            src = f"./src/{self.dir_name}"
        else:
            src = f"./{self.dir_name}"
        return src

    @property
    def precompiled_globs(self):
        _globs = []
        for ex in self.precompiled_extension:
            _globs.extend((f"{self.project_dir}/*{ex}", f"{self.project_dir}/**/*{ex}"))
        return list(set(_globs))

    @property
    @memo
    def options_exclude(self):
        return [parse_user_glob(e) for e in self.options.files.exclude]

    def filter_ensure_wanted(self, tgts: ListStr):
        return list(
            filter(
                lambda s: not any(re.match(e, self.normalize_glob(s), re.IGNORECASE) for e in self.options_exclude),
                tgts,
            )
        )

    @property
    @memo
    def included_files(self):
        included = []
        _normu = [self.normalize_glob(parse_user_glob(e)) for e in self.options.files.exclude]
        self.app.display_debug("user globs")
        self.app.display_debug(_normu)
        for patt in self.precompiled_globs:
            globbed = glob(patt, recursive=True)
            if len(globbed) == 0:
                continue
            matched = self.filter_ensure_wanted(globbed)
            included.extend(matched)
        return included

    @property
    @memo
    def normalized_included_files(self):
        """
        Produces files in posix format
        """
        return [self.normalize_glob(f) for f in self.included_files]

    @property
    @memo
    def grouped_included_files(self) -> ListT[ExtensionArg]:
        grouped = {}
        for norm in self.normalized_included_files:
            root, ext = os.path.splitext(norm)
            ok = True
            if ext == ".pxd":
                pyfile = norm.replace(".pxd", ".py")
                if os.path.exists(pyfile):
                    norm = pyfile  # noqa: PLW2901
                else:
                    ok = False
                    self.app.display_warning(f"attempted to use .pxd file without .py file ({norm})")
            if self.is_src:
                root = root.replace("./src/", "")
            root = root.replace("/", ".")
            alias = self.options.files.matches_alias(root)
            if alias:
                root = alias
            if grouped.get(root) and ok:
                grouped[root].append(norm)
            elif ok:
                grouped[root] = [norm]
        return [ExtensionArg(name=key, files=files) for key, files in grouped.items()]

    @property
    @memo
    def artifact_globs(self):
        artifact_globs = []
        for included_file in self.normalized_included_files:
            root, _ = os.path.splitext(included_file)
            artifact_globs.extend(f"{root}.*{ext}" for ext in self.precompiled_extension)
        return artifact_globs

    @property
    @memo
    def normalized_artifact_globs(self):
        """
        Produces files in platform native format (e.g. a/b vs a\\b)
        """
        return [self.normalize_glob(f) for f in self.artifact_globs]

    @property
    @memo
    def artifact_patterns(self):
        return [f"/{artifact_glob}" for artifact_glob in self.normalized_artifact_globs]

    @contextmanager
    def get_build_dirs(self):
        with TemporaryDirectory() as temp_dir:
            yield os.path.realpath(temp_dir)

    def _globs(self, exts: ListStr):
        globs = [
            *(f"{self.project_dir}/**/*{ext}" for ext in exts),
            *(f"{self.project_dir}/*{ext}" for ext in exts),
        ]
        globbed = []
        for g in globs:
            globbed += [self.normalize_path(f) for f in glob(g, recursive=True)]
        return list(set(globbed))

    @property
    def intermediate(self):
        return self._globs(self.intermediate_extensions)

    @property
    def compiled(self):
        return self._globs(self.compiled_extensions)

    @property
    def inclusion_map(self):
        include = {}
        for compl in self.compiled:
            include[compl] = compl
        self.app.display_debug("Derived inclusion map")
        self.app.display_debug(include)
        return include

    def rm_recurse(self, li: ListStr):
        self.app.display_debug("Removing by match")
        self.app.display_debug(li)
        for f in li:
            os.remove(f)

    def clean_intermediate(self):
        self.rm_recurse(self.intermediate)

    def clean_compiled(self):
        self.rm_recurse(self.compiled)

    def clean(self, _versions: ListStr):
        self.clean_intermediate()
        self.clean_compiled()

    @property
    @memo
    def options(self):
        config = parse_from_dict(self)
        if config.compile_py:
            self.precompiled_extension.append(".py")
        return config

    def initialize(self, _: str, build_data: dict):
        self.app.display_mini_header(self.PLUGIN_NAME)
        self.app.display_debug("Options")
        self.app.display_debug(self.options.asdict(), level=1)
        self.app.display_waiting("Pre-build artifacts")
        self.app.display_info("Building c/c++ extensions...")
        self.app.display_info(self.normalized_included_files)
        with self.get_build_dirs() as temp:
            shared_temp_build_dir = os.path.join(temp, "build")
            temp_build_dir = os.path.join(temp, "tmp")
            os.mkdir(shared_temp_build_dir)
            os.mkdir(temp_build_dir)
            setup_file = os.path.join(temp, "setup.py")
            with open(setup_file, "w") as f:
                setup = setup_py(
                    *self.grouped_included_files,
                    options=self.options,
                )
                self.app.display_debug(setup)
                f.write(setup)

            self.options.validate_include_opts()

            process = subprocess.run(  # noqa: PLW1510
                [  # noqa: S603
                    sys.executable,
                    setup_file,
                    "build_ext",
                    "--inplace",
                    "--verbose",
                    "--build-lib",
                    shared_temp_build_dir,
                    "--build-temp",
                    temp_build_dir,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                env=self.options.envflags.env,
            )
            stdout = process.stdout.decode("utf-8")
            if process.returncode:
                self.app.display_error(f"cythonize exited non null status {process.returncode}")
                self.app.display_error(stdout)
                msg = "failed compilation"
                raise Exception(msg)
            else:
                self.app.display_info(stdout)

            self.app.display_success("Post-build artifacts")
            self.app.display_info(glob(f"{self.project_dir}/*/**", recursive=True))

        if not self.options.retain_intermediate_artifacts:
            self.clean_intermediate()

        build_data["infer_tag"] = True
        build_data["pure_python"] = False
        build_data["artifacts"].extend(self.artifact_patterns)
        build_data["force_include"].update(self.inclusion_map)
        self.app.display_info("Extensions complete")



from dataclasses import dataclass, field


@dataclass
class Autoimport:
    pkg: str

    include: str
    libraries: str = field(default=None)
    library_dirs: str = field(default=None)
    required_call: str = field(default=None)


__packages__ = {
    a.pkg: a
    for a in (
        Autoimport("numpy", "get_include"),
        Autoimport(
            "pyarrow",
            include="get_include",
            libraries="get_libraries",
            library_dirs="get_library_dirs",
            required_call="create_library_symlinks",
        ),
    )
}



# SPDX-FileCopyrightText: 2023-present joshua-auchincloss <joshua.auchincloss@proton.me>
#
# SPDX-License-Identifier: MIT
import os
import sys

src = os.sep.join(__file__.split(os.sep)[:-2])
sys.path.append(src)

from hatch_cython import CythonBuildHook  # noqa: E402, F401



from typing import Optional


class ValueDefn(object):  # noqa: UP004
    def __init__(self, value: Optional[int] = None):
        self.value = value if value else 0

    def set(self, value):  # noqa: A003
        v = self.value
        self.value = value
        if v:
            return True
        return False



from collections.abc import Generator
from dataclasses import asdict, dataclass, field
from importlib import import_module
from os import path
from typing import Optional

from hatch.utils.ci import running_in_ci
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from hatch_cython.config.autoimport import Autoimport
from hatch_cython.config.defaults import get_default_compile, get_default_link
from hatch_cython.config.files import FileArgs
from hatch_cython.config.flags import EnvFlags, parse_env_args
from hatch_cython.config.includes import parse_includes
from hatch_cython.config.macros import DefineMacros, parse_macros
from hatch_cython.config.platform import ListedArgs, PlatformArgs, parse_platform_args
from hatch_cython.constants import DIRECTIVES, EXIST_TRIM, INCLUDE, LTPY311, MUST_UNIQUE
from hatch_cython.types import CallableT, ListStr

# fields tracked by this plugin
__known__ = (
    "src",
    "env",
    "files",
    "compile_py",
    "includes",
    "libraries",
    "library_dirs",
    "define_macros",
    "directives",
    "compile_args",
    "cythonize_kwargs",
    "extra_link_args",
    "retain_intermediate_artifacts",
)


def parse_from_dict(cls: BuildHookInterface):
    given = cls.config.get("options", {})
    passed = given.copy()
    kwargs = {}
    for kw, val in given.items():
        if kw in __known__:
            parsed: any
            if kw == "files":
                val: dict
                parsed: FileArgs = FileArgs(**val)
            elif kw == "define_macros":
                val: list
                parsed: DefineMacros = parse_macros(val)
            else:
                val: any
                parsed: any = val
            kwargs[kw] = parsed
            passed.pop(kw)
            continue

    compile_args = parse_platform_args(kwargs, "compile_args", get_default_compile)
    link_args = parse_platform_args(kwargs, "extra_link_args", get_default_link)
    envflags = parse_env_args(kwargs)
    cfg = Config(**kwargs, compile_args=compile_args, extra_link_args=link_args, envflags=envflags)

    for kw, val in passed.copy().items():
        is_include = kw.startswith(INCLUDE)
        if is_include and val:
            cfg.resolve_pkg(
                cls,
                parse_includes(kw, val),
            )
            passed.pop(kw)
            continue
        elif is_include:
            passed.pop(kw)
            continue
        elif kw == "parallel" and passed.get(kw):
            comp = [
                PlatformArgs(arg="/openmp", platforms="windows"),
                PlatformArgs(arg="-fopenmp", platforms=["linux"]),
            ]
            link = [
                PlatformArgs(arg="/openmp", platforms="windows"),
                PlatformArgs(arg="-fopenmp", platforms="linux"),
                PlatformArgs(arg="-lomp", platforms="darwin", marker=LTPY311, apply_to_marker=running_in_ci),
            ]
            cma = ({*cfg.compile_args}).union({*comp})
            cfg.compile_args = list(cma)
            seb = ({*cfg.extra_link_args}).union({*link})
            cfg.extra_link_args = list(seb)
            passed.pop(kw)

    cfg.compile_kwargs = passed
    return cfg


@dataclass
class Config:
    src: Optional[str] = field(default=None)  # noqa: UP007
    files: FileArgs = field(default_factory=FileArgs)
    includes: ListStr = field(default_factory=list)
    define_macros: DefineMacros = field(default_factory=list)
    libraries: ListStr = field(default_factory=list)
    library_dirs: ListStr = field(default_factory=list)
    directives: dict = field(default_factory=lambda: DIRECTIVES)
    compile_args: ListedArgs = field(default_factory=get_default_compile)
    compile_kwargs: dict = field(default_factory=dict)
    cythonize_kwargs: dict = field(default_factory=dict)
    extra_link_args: ListedArgs = field(default_factory=get_default_link)
    retain_intermediate_artifacts: bool = field(default=False)
    envflags: EnvFlags = field(default_factory=EnvFlags)
    compile_py: bool = field(default=True)

    def __post_init__(self):
        self.directives = {**DIRECTIVES, **self.directives}

    @property
    def compile_args_for_platform(self):
        return self._arg_impl(self.compile_args)

    @property
    def compile_links_for_platform(self):
        return self._arg_impl(self.extra_link_args)

    def _post_import_attr(
        self,
        cls: BuildHookInterface,
        im: Autoimport,
        att: str,
        mod: any,
        extend: CallableT[[ListStr], None],
        append: CallableT[[str], None],
    ):
        attr = getattr(im, att)
        if attr is not None:
            try:
                libraries = getattr(mod, attr)
                if callable(libraries):
                    libraries = libraries()

                if isinstance(libraries, str):
                    append(libraries)
                elif isinstance(libraries, (list, Generator)):  # noqa: UP038
                    extend(libraries)
                elif isinstance(libraries, dict):
                    extend(libraries.values())
                else:
                    cls.app.display_warning(f"{im.pkg}.{attr} has an invalid type ({type(libraries)})")

            except AttributeError:
                cls.app.display_warning(f"{im.pkg}.{attr}")

    def resolve_pkg(
        self,
        cls: BuildHookInterface,
        im: Autoimport,
    ):
        mod = import_module(im.pkg)
        calls = getattr(mod, im.include)
        if not callable(calls):
            msg = f"{im.pkg}.{im.include} is invalid"
            raise ValueError(msg)
        self.includes.append(calls())
        self._post_import_attr(
            cls,
            im,
            "libraries",
            mod,
            self.libraries.extend,
            self.libraries.append,
        )
        self._post_import_attr(
            cls,
            im,
            "library_dirs",
            mod,
            self.library_dirs.extend,
            self.library_dirs.append,
        )
        if im.required_call is not None:
            if hasattr(mod, im.required_call):
                call = getattr(mod, im.required_call)
                call()
            else:
                cls.app.display_warning(f"{im.pkg}.{im.required_call} is invalid")

    def _arg_impl(self, target: ListedArgs):
        args = {"any": []}

        def with_argvalue(arg: str):
            # be careful with e.g. -Ox flags
            matched = list(filter(lambda s: arg.startswith(s), MUST_UNIQUE))
            if len(matched):
                m = matched[0]
                args[m] = arg.split(" ")
            else:
                args["any"].append(arg.split(" "))

        for arg in target:
            # if compile-arg format, check platform applies
            if isinstance(arg, PlatformArgs):
                if arg.applies() and arg.is_exist(EXIST_TRIM):
                    with_argvalue(arg.arg)
            # else assume string / user knows what theyre doing and add to the call params
            else:
                with_argvalue(arg)

        flat = []

        def flush(it):
            if isinstance(it, list):
                for v in it:
                    flush(v)
            else:
                flat.append(it)

        # side effect
        list(map(flush, args.values()))
        return flat

    def asdict(self):
        d = asdict(self)
        d["envflags"]["env"] = self.envflags.masked_environ()
        return d

    def validate_include_opts(self):
        for opt in self.includes:
            if not path.exists(opt):
                msg = "%s does not exist" % opt
                raise ValueError(msg)



# SPDX-FileCopyrightText: 2023-present joshua-auchincloss <joshua.auchincloss@proton.me>
#
# SPDX-License-Identifier: MIT
from hatch_cython.hooks import hatch_register_build_hook
from hatch_cython.plugin import CythonBuildHook



from collections.abc import Hashable
from dataclasses import dataclass
from os import path

from packaging.markers import Marker

from hatch_cython.constants import ANON
from hatch_cython.types import CallableT, ListStr, ListT, UnionT
from hatch_cython.utils import aarch, plat


@dataclass
class PlatformBase(Hashable):
    platforms: UnionT[ListStr, str] = "*"
    arch: UnionT[ListStr, str] = "*"
    depends_path: bool = False
    marker: str = None
    apply_to_marker: CallableT[[], bool] = None

    def __post_init__(self):
        self.do_rewrite("platforms")
        self.do_rewrite("arch")

    def do_rewrite(self, attr: str):
        att = getattr(self, attr)
        if isinstance(att, list):
            setattr(self, attr, [p.lower() for p in att])
        elif isinstance(att, str):
            setattr(self, attr, att.lower())

    def check_marker(self):
        do = True
        if self.apply_to_marker:
            do = self.apply_to_marker()
        if do:
            marker = Marker(self.marker)
            return marker.evaluate()
        return False

    def _applies_impl(self, attr: str, defn: str):
        if self.marker:
            ok = self.check_marker()
            if not ok:
                return False

        att = getattr(self, attr)
        if isinstance(att, list):
            # https://docs.python.org/3/library/platform.html#platform.machine
            # "" is a possible value so we have to add conditions for anon
            _anon = ANON in att and defn == ""
            return defn in att or "*" in att or _anon
        _anon = ANON == att and defn == ""
        return (att in (defn, "*")) or _anon

    def applies(self):
        _isplatform = self._applies_impl("platforms", plat())
        _isarch = self._applies_impl("arch", aarch())
        return _isplatform and _isarch

    def is_exist(self, trim: int = 0):
        if self.depends_path:
            return path.exists(self.arg[trim:])
        return True


@dataclass
class PlatformArgs(PlatformBase):
    arg: str = None

    def __hash__(self) -> int:
        return hash(self.arg)


def parse_to_plat(cls, arg, args: UnionT[list, dict], key: UnionT[int, str], require_argform: bool, **kwargs):
    if isinstance(arg, dict):
        args[key] = cls(**arg, **kwargs)
    elif require_argform:
        msg = f"arg {key} is invalid. must be of type ({{ flag = ... , platform = '*' }}) given {arg} ({type(arg)})"
        raise ValueError(msg)


def parse_platform_args(
    kwargs: dict,
    name: str,
    default: CallableT[[], ListT[PlatformArgs]],
) -> ListT[UnionT[str, PlatformArgs]]:
    try:
        args = [*default(), *kwargs.pop(name)]
        for i, arg in enumerate(args):
            parse_to_plat(PlatformArgs, arg, args, i, require_argform=False)
    except KeyError:
        args = default()
    return args


ListedArgs = ListT[UnionT[PlatformArgs, str]]
"""
List[str | PlatformArgs]
"""



from pytest import raises

from hatch_cython.config.autoimport import Autoimport
from hatch_cython.config.includes import parse_includes


def test_includes():
    parse = {
        "pkg": "somelib",
        "include": "gets_include",
        "libraries": "gets_libraries",
        "library_dirs": "gets_library_dirs",
        "required_call": "some_setup_op",
    }
    ok = parse_includes(
        "include_abc",
        parse,
    )
    assert ok == Autoimport(**parse)
    ok = parse_includes(
        "include_abc",
        "some_attr",
    )
    assert ok == Autoimport(pkg="abc", include="some_attr")


def test_invalid():
    with raises(ValueError, match="either provide a known package"):
        parse_includes("list", [])



from dataclasses import dataclass, field
from os import environ
from typing import ClassVar

from hatch_cython.config.platform import PlatformArgs, parse_to_plat
from hatch_cython.types import CallableT, DictT
from hatch_cython.utils import path_delim


@dataclass
class EnvFlag(PlatformArgs):
    env: str = field(default="")
    merges: bool = field(default=False)
    sep: str = field(default=" ")

    def __hash__(self) -> int:
        return hash(self.field)


__flags__ = (
    EnvFlag(env="CC", merges=False),
    EnvFlag(env="CPP", merges=False),
    EnvFlag(env="CXX", merges=False),
    EnvFlag(env="CFLAGS", merges=True),
    EnvFlag(env="CCSHARED", merges=True),
    EnvFlag(env="CPPFLAGS", merges=True),
    EnvFlag(env="LDFLAGS", merges=True),
    EnvFlag(env="LDSHARED", merges=True),
    EnvFlag(env="SHLIB_SUFFIX", merges=False),
    EnvFlag(env="AR", merges=False),
    EnvFlag(env="ARFLAGS", merges=True),
    EnvFlag(env="PATH", merges=True, sep=path_delim()),
)


@dataclass
class EnvFlags:
    CC: PlatformArgs = None
    CPP: PlatformArgs = None
    CXX: PlatformArgs = None

    CFLAGS: PlatformArgs = None
    CCSHARED: PlatformArgs = None

    CPPFLAGS: PlatformArgs = None

    LDFLAGS: PlatformArgs = None
    LDSHARED: PlatformArgs = None

    SHLIB_SUFFIX: PlatformArgs = None

    AR: PlatformArgs = None
    ARFLAGS: PlatformArgs = None

    PATH: PlatformArgs = None

    custom: DictT[str, PlatformArgs] = field(default_factory=dict)
    env: dict = field(default_factory=environ.copy)

    __known__: ClassVar[DictT[str, EnvFlag]] = {e.env: e for e in __flags__}

    def __post_init__(self):
        for flag in __flags__:
            self.merge_to_env(flag, self.get_from_self)
        for flag in self.custom.values():
            self.merge_to_env(flag, self.get_from_custom)

    def merge_to_env(self, flag: EnvFlag, get: CallableT[[str], EnvFlag]):
        var = environ.get(flag.env)
        override: EnvFlag = get(flag.env)
        if override and flag.merges:
            add = var + flag.sep if var else ""
            self.env[flag.env] = add + override.arg
        elif override:
            self.env[flag.env] = override.arg

    def get_from_self(self, attr):
        return getattr(self, attr)

    def get_from_custom(self, attr):
        return self.custom.get(attr)

    def masked_environ(self) -> dict:
        out = {}
        for k, v in self.env.items():
            if k not in self.__known__:
                out[k] = "*" * len(v)
            else:
                out[k] = v
        return out


def parse_env_args(
    kwargs: dict,
) -> EnvFlags:
    try:
        args: list = kwargs.pop("env")
        for i, arg in enumerate(args):
            parse_to_plat(EnvFlag, arg, args, i, require_argform=True)
    except KeyError:
        args = []
    kw = {"custom": {}}
    for arg in args:
        arg: EnvFlag
        if arg.applies():
            if arg.env in EnvFlags.__known__:
                kw[arg.env] = arg
            else:
                kw["custom"][arg.env] = arg
    envflags = EnvFlags(**kw)
    return envflags



import subprocess
import sys
from glob import glob
from logging import getLogger

logger = getLogger()

if __name__ == "__main__":
    artifact = glob("dist/example*.whl")[0]
    proc = subprocess.run(  # noqa: PLW1510
        [sys.executable, "-m", "pip", "install", artifact, "--force-reinstall"],  # noqa: S603
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    if proc.returncode:
        logger.error(proc.stdout.decode())
        raise SystemExit(1)



from hatch_cython.config.autoimport import Autoimport, __packages__
from hatch_cython.constants import INCLUDE


def parse_includes(kw: str, val: str):
    alias = kw.replace(INCLUDE, "")
    import_p = __packages__.get(alias)
    if import_p is None:
        if isinstance(val, str):
            import_p = Autoimport(pkg=alias, include=val)
        elif isinstance(val, dict):
            if "pkg" not in val:
                val["pkg"] = alias
            import_p = Autoimport(**val)
        else:
            msg = " ".join(
                (
                    "%s (%s) is invalid, either provide a known package or",
                    "a path in the format of module.get_xxx where get_xxx is",
                    "the directory to be included",
                )
            ).format(val, type(val))
            raise ValueError(msg)
    return import_p
